#
# Copyright (C) 2016 VSCT  
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License.  
# You may obtain a copy of the License at 
#
# http://www.apache.org/licenses/LICENSE-2.0  
#
# Unless required by applicable law or agreed to in writing, software  
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License. 
#

import requests


def aslist(generator):
    "Function decorator to transform a generator into a list"
    def wrapper(*args, **kwargs):
        return list(generator(*args, **kwargs))
    return wrapper



#
# Copyright (C) 2016 VSCT  
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License.  
# You may obtain a copy of the License at 
#
# http://www.apache.org/licenses/LICENSE-2.0  
#
# Unless required by applicable law or agreed to in writing, software  
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License. 
#

import requests
from distutils.command.upload import upload as PypiUploadCmd
from distutils.errors import DistutilsOptionError

# Documentation for this pattern: https://docs.python.org/3/distutils/extending.html
class NexusUploadCmd(PypiUploadCmd):
    user_options = PypiUploadCmd.user_options + [
        ('username=', None, 'HTTP BasicAuth username to upload to your Sonatype Nexus'),
        ('password=', None, 'HTTP BasicAuth password to upload to your Sonatype Nexus'),
    ]

    def upload_file(self, command, pyversion, filename): # pylint: disable=unused-argument
        if not self.username or not self.password:
            raise DistutilsOptionError('--username and --password parameters are both required, either passed as CLI parameters or defined in .pypirc')
        extension = 'tar.gz' if filename.endswith('.tar.gz') else filename.rpartition('.')[-1]
        platforms = self.distribution.get_platforms()
        if platforms == ['UNKNOWN']:
            platforms = ['py2.py3-none-any']
        for platform in platforms:
            artifact_url = '{base_url}/{name}/{version}/{name}-{version}-{classifier}.{extension}'.format(
                    name=self.distribution.get_name(), version=self.distribution.get_version(),
                    base_url=self.repository, classifier=platform, extension=extension)
            print('Uploading to', artifact_url)
            with open(filename, 'rb') as artifact_binary:
                response = requests.post(artifact_url, data=artifact_binary, auth=(self.username, self.password))
                if 400 <= response.status_code < 600:
                    raise requests.HTTPError(str(response.status_code) + '\n' + response.text)



#
# Copyright (C) 2016 VSCT  
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License.  
# You may obtain a copy of the License at 
#
# http://www.apache.org/licenses/LICENSE-2.0  
#
# Unless required by applicable law or agreed to in writing, software  
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License. 
#

"Code to interact with Pypi"

import requests


class PypiQueryError(Exception):
    pass

def get_package_release_from_pypi(pkg_name, version):
    """
    No classifier-based selection of Python packages is currently implemented: for now we don't fetch any .whl or .egg
    Eventually, we should select the best release available, based on the classifier & PEP 425.
    E.g. a wheel when available but NOT for tornado 4.3 for example, where available wheels are only for Windows.
    Note also that some packages don't have .whl distributed, e.g. https://bugs.launchpad.net/lxml/+bug/1176147
    """
    matching_releases = get_package_releases_matching_version(pkg_name, version)
    src_releases = [release for release in matching_releases if release['python_version'] == 'source']
    if not src_releases:
        raise PypiQueryError('No source distribution found for package {} version {}'.format(pkg_name, version))
    for release in src_releases:
        release['classifier'], release['extension'] = extract_classifier_and_extension(pkg_name, release['filename'])
    try:  # we return the release with the most generic classifier if there is one
        return next(release for release in src_releases if release['classifier'] == 'py2.py3-none-any')
    except StopIteration:
        return src_releases[0]

def get_package_releases_matching_version(pkg_name, version):
    releases = get_package_releases(pkg_name)
    if version not in releases:
        raise PypiQueryError('Version {} of package {} not found in Pypi. Available versions: {}', version, pkg_name, ','.join(releases.keys()))
    return releases[version]

def get_package_releases(pkg_name):
    response = requests.get('http://pypi.python.org/pypi/{}/json'.format(pkg_name))
    response.raise_for_status()
    return response.json()['releases']

def extract_classifier_and_extension(pkg_name, filename):
    """
    Returns a PEP425-compliant classifier (or 'py2.py3-none-any' if it cannot be extracted),
    and the file extension
    """
    basename, _, extension = filename.rpartition('.')
    if extension == 'gz' and filename.endswith('.tar.gz'):
        extension = 'tar.gz'
        basename = filename[:-7]
    if basename == pkg_name or basename[len(pkg_name)] != '-':
        return 'py2.py3-none-any', extension
    basename = basename[len(pkg_name)+1:]
    classifier_parts = basename.split('-')
    if len(classifier_parts) < 3:
        return 'py2.py3-none-any', extension
    if len(classifier_parts) == 3:
        _, _, classifier_parts[0] = classifier_parts[0].rpartition('.')
    return '-'.join(classifier_parts[-3:]), extension




#
# Copyright (C) 2016 VSCT  
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License.  
# You may obtain a copy of the License at 
#
# http://www.apache.org/licenses/LICENSE-2.0  
#
# Unless required by applicable law or agreed to in writing, software  
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License. 
#

"Code to interact with Sonatype Nexus"

import logging, re, requests

from .pypi import get_package_release_from_pypi
from .utils import aslist


logger = logging.getLogger(__name__)


class NexusRepositoryClient:
    """
    Heavily inspired by stardust85/repositorytools repository.NexusRepositoryClient._upload_artifact
    API details: https://repository.sonatype.org/nexus-restlet1x-plugin/default/docs/rest.html
    """
    def __init__(self, base_url, repo_id, artifact_group, auth):
        self.base_url = base_url
        self.repo_id = repo_id
        self.artifact_group = artifact_group
        self.auth = auth
        self._session = requests.session()

    def upload_from_pypi_if_need_be(self, pkg_name, pkg_version):
        logger.debug('Checking for Python package %s version %s in Nexus', pkg_name, pkg_version)
        matching_artifacts = self.find_artifacts(pkg_name, pkg_version)
        if matching_artifacts:
            logger.info('Skipping Python package %s version %s - the following artifacts are already in Nexus : %s',
                pkg_name, pkg_version, ','.join(url.split('/')[-1] for url in matching_artifacts))
            if len(matching_artifacts) > 1:
                raise NotImplementedError('Classifier-based selection of Python packages is not implemented yet.\n'
                                          'Several matching artifacts found in Nexus: {}'.format(matching_artifacts))
            return matching_artifacts[0]
        else:
            logger.debug('Retrieving release for Python package %s version %s from Pypi', pkg_name, pkg_version)
            pkg_release = get_package_release_from_pypi(pkg_name, pkg_version)

            response = requests.get(pkg_release['url'])
            response.raise_for_status()
            pkg_binary = response.content

            logger.debug('Uploading Python package %s version %s to Nexus', pkg_name, pkg_version)
            artifact_url = self.get_artifact_url(pkg_name, pkg_version, pkg_release['classifier'], pkg_release['extension'])
            self.upload_artifact(artifact_url, pkg_binary)
            return artifact_url

    @staticmethod
    def get_artifact_filename(name, version, classifier, extension):
        return '{name}-{version}-{classifier}.{extension}'.format(name=name, version=version, classifier=classifier, extension=extension)

    def get_artifact_url_dir(self, name, version):
        return '{hostname}/content/repositories/{repo_id}/{group}/{name}/{version}'.format(
            hostname=self.base_url, repo_id=self.repo_id, group=self.artifact_group, name=name, version=version)

    def get_artifact_url(self, pkg_name, version, classifier, extension):
        artifact_nexus_filename = self.get_artifact_filename(pkg_name, version, classifier, extension)
        return '{}/{}'.format(self.get_artifact_url_dir(pkg_name, version), artifact_nexus_filename)

    def find_artifacts(self, name, version):
        url = self.get_artifact_url_dir(name, version)
        response = self._session.get(url)
        if response.status_code == 404:
            return []
        response.raise_for_status()
        return self._extract_artifacts_url(url, response.text)

    @staticmethod
    def pkg_name_version_from_artifact_url(artifact_url):
        "Returns a (pkg_name, pkg_version) tuple"
        return artifact_url.split('/')[-3:-1]

    def append_egg_hash_if_nexus_url(self, url):
        if url.startswith(self.base_url) and '#egg=' not in url:
            return url + '#egg={}=={}'.format(*self.pkg_name_version_from_artifact_url(url))
        return None

    @staticmethod
    @aslist
    def _extract_artifacts_url(base_url, nexus_html_listing):
        # OK, regex-parsing HTML is uglyyy, but there is no need to introduce an external dependency to beautifulsoup4 just for this
        for url in re.findall(r' href="(' + base_url + r'/[^"]+)"', nexus_html_listing):
            if not (url.endswith('.pom') or url.endswith('.md5') or url.endswith('.sha1')):
                yield url

    def upload_artifact(self, artifact_url, artifact_binary):
        self._session.post(artifact_url, data=artifact_binary, auth=self.auth).raise_for_status()



#!/usr/bin/env python3

#
# Copyright (C) 2016 VSCT  
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License.  
# You may obtain a copy of the License at 
#
# http://www.apache.org/licenses/LICENSE-2.0  
#
# Unless required by applicable law or agreed to in writing, software  
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License. 
#

# INSTALL: pip install pip-tools requests (the former is needed to get the pip-compile command)
# AUTHOR: Lucas Cimon

import argparse, getpass, logging, os, sys
from urllib.parse import urlparse
from .nexus import NexusRepositoryClient
from .pip_compile import pip_compile
from .requirements import read_requirements, requirements2reqfile_lines, subst_editable_pkg_fallback, filter_out_provided_requirements


def main(argv=sys.argv[1:]):
    args = parse_args(argv)
    if args.debug:
        logging.basicConfig(level=logging.DEBUG, stream=sys.stderr,
            format="%(asctime)s - pid:%(process)s %(filename)s:%(lineno)d [%(levelname)s] %(message)s")
    else:
        logging.basicConfig(level=logging.INFO, stream=sys.stderr, format="%(message)s")
    nexus_client = NexusRepositoryClient(base_url=args.nexus_url, repo_id=args.repo_id, artifact_group=args.artifact_group.replace('.', '/'), auth=(args.user, args.password))
    with open(args.in_reqfile, 'r') as in_reqfile_content:
        requirements = read_requirements(in_reqfile_content.readlines())
    requirements = [subst_editable_pkg_fallback(req) for req in requirements]
    requirements = read_requirements(pip_compile(requirements2reqfile_lines(requirements), urlparse(nexus_client.base_url).netloc, nexus_client.append_egg_hash_if_nexus_url))
    if args.default_packages:
        with open(args.default_packages, 'r') as default_pkgs_file:
            requirements = filter_out_provided_requirements(requirements, read_requirements(default_pkgs_file.readlines()))
    with open(args.out_reqfile, 'w') as out_reqfile_content:
        for req_type, req_info in requirements:
            if req_type == 'url':
                requirement_url, comment = req_info
                out_reqfile_content.write(requirement_url + (' # ' + comment if comment else '') + '\n')
            elif req_type == 'version-locked':
                pkg_name, pkg_version, comment = req_info
                nexus_artifact_url = nexus_client.upload_from_pypi_if_need_be(pkg_name, pkg_version)
                out_reqfile_content.write(nexus_artifact_url + (' # ' + comment if comment else '') + '\n')

def parse_args(argv):
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--nexus-url', required=True, help=' ')
    parser.add_argument('--artifact-group', required=True, help=' ')
    parser.add_argument('--repo-id', default='pip', help=' ')
    parser.add_argument('--default-packages', help='File containing a requirements.txt-style list of Python packages already included in the base environment')
    parser.add_argument('--user', help='A prompt will ask for your password - Else use environment variables REPOSITORY_USER / REPOSITORY_PASSWORD')
    parser.add_argument('--in-reqfile', default='requirements.txt')
    parser.add_argument('--out-reqfile', default='nexus-requirements.txt')
    # Subparsers could easily be added here, to enumerate artifacts / only retrieve a package release from Pypi / onlyu upload an artifact to nexus... but YAGNI prevails
    args = parser.parse_args(argv)
    if args.user:
        args.password = getpass.getpass(prompt='Please enter password for user {} : '.format(args.user))
    else:
        if 'REPOSITORY_USER' not in os.environ or 'REPOSITORY_PASSWORD' not in os.environ:
            parser.error('Repository user not specified. Please use the interactive --user option or define the environment variables REPOSITORY_USER / REPOSITORY_PASSWORD')
        args.user = os.environ['REPOSITORY_USER']
        args.password = os.environ['REPOSITORY_PASSWORD']
    return args

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os, sys
from setuptools import setup, find_packages

version = '1.0.2'
setup_requires = []

if 'nexus_upload' in sys.argv:
    sys.argv.remove('nexus_upload')
    setup_requires.append('nexus_uploader')

if '--snapshot' in sys.argv:
    sys.argv.remove('--snapshot')
    version += '-SNAPSHOT'

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')) as file_requirements:
    requirements = file_requirements.readlines()

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = open('README.md').read()

setup(
    name='nexus_uploader',
    description='CLI tool to upload Python packages listed in a requirements.txt file into a Sonatype Nexus (from Pypi), and return the list of the artifact URLs',
    long_description=long_description,
    author='Lucas Cimon',
    url='http://github.com/voyages-sncf-technologies/nexus_uploader',
    install_requires=requirements,
    packages=find_packages(),
    version=version,
    setup_requires=setup_requires,
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
    ],
    keywords='nexus artifacts packages upload requirements pip-compile',
    license='Apache License 2.0',
    include_package_data=True,  # Active la prise en compte du fichier MANIFEST.in
    entry_points={
        'console_scripts': ['pyRequirements2nexus = nexus_uploader.pyrequirements2nexus:main'],
        'distutils.commands': ['nexus_upload = nexus_uploader.nexus_upload_cmd:NexusUploadCmd'],
    },
)
